#include <algorithm>

#include "Filter.hpp" 

Filter::Filter()
{
    filter_type = no_filter;;
    reset_kernel();
    edit_number_of_packet(1);
}

void Filter::setUp(std::shared_ptr<ThreadPool> iThreadPool)
{
    mThreadPool = iThreadPool;
}

void Filter::print()
{
    switch (filter_type)
    {
        case no_filter:
            std::cout << "No filter apply" << std::endl;
            break;
        case convolution:
            std::cout << "Convolution filter apply" << std::endl;
            std::cout << "Normalised convolution kernel: " << mKernel;
            break;
        case median:
            std::cout << "Median filter apply" << std::endl;
            break;
        default:
            std::cout << "Warning, wrong filter parameter..." << std::endl;
            break;
    }
}

void Filter::edit_filter_type(FilterType nfilter_type)
{
    if ((nfilter_type < 0) || (nfilter_type > 2))
    {
        std::cout << "Warning, wrong filter parameter" << std::endl;
        nfilter_type = no_filter;
    }
    filter_type = nfilter_type;
}

//Convolution filter
void Filter::reset_kernel()
{
    if ((mKernel.row_size() != 3) || (mKernel.column_size() !=3)) {mKernel.resize(3,3);}
    for (int i=0 ; i<3 ; i++)
    {
        for (int j=0 ; j<3 ; j++) {mKernel.data[i][j]=1.f;}
    }
    normalise_kernel();
}

void Filter::edit_kernel(image<float> iKernel)
{
    if (iKernel.sum() < 1.f/1000000.f)
    {
        //better normalisation should be done such as max/normalisation < x
        std::cout << "Warning, sum of element is too small for normalisation" << std::endl;
        if (mKernel.sum() < 1.f/1000000.f) {reset_kernel();}
    }
    else
    {
        mKernel=iKernel;
        normalisation = mKernel.sum();
        normalise_kernel();
    }
}

void Filter::edit_kernel(float f00, float f01, float f02, float f10, float f11, float f12, float f20, float f21, float f22)
{
    image<float> lim(3,3);
    lim.data[0][0]=f00;
    lim.data[0][1]=f01;
    lim.data[0][2]=f02;
    lim.data[1][0]=f10;
    lim.data[1][1]=f11;
    lim.data[1][2]=f12;
    lim.data[2][0]=f20;
    lim.data[2][1]=f21;
    lim.data[2][2]=f22;
    edit_kernel(lim);
}

void Filter::normalise_kernel()
{
    normalisation = mKernel.sum();
    for(int i = 0; i < 3; i++)
    {
        for(int j = 0; j < 3; j++) {mKernel.data[i][j] = mKernel.data[i][j] / normalisation;}
    }
    normalisation2x3top = mKernel.data[1][0]+mKernel.data[1][1]+mKernel.data[1][2] + mKernel.data[2][0]+mKernel.data[2][1]+mKernel.data[2][2];
    normalisation2x3bottom = mKernel.data[0][0]+mKernel.data[0][1]+mKernel.data[0][2] + mKernel.data[1][0]+mKernel.data[1][1]+mKernel.data[1][2];
    normalisation3x2left = mKernel.data[0][1]+mKernel.data[1][1]+mKernel.data[2][1] + mKernel.data[0][2]+mKernel.data[1][2]+mKernel.data[2][2];
    normalisation3x2right = mKernel.data[0][0]+mKernel.data[1][0]+mKernel.data[2][0] + mKernel.data[0][1]+mKernel.data[1][1]+mKernel.data[2][1];
}

//Median filter



//General filter funtionc
void Filter::NoFilter(const image<float> &iImageInput, image<float> &oImageOutput)
{
    int lImageWidth = iImageInput.column_size();
    int lImageHeight = iImageInput.row_size();

    if(lImageWidth != oImageOutput.column_size() || lImageHeight != oImageOutput.row_size())
    {
        std::cerr << "Image data format invalid - Input/Output" << std::endl;
        return;
    }
    
    oImageOutput=iImageInput;
}

void Filter::applyConvolutionFilter(const image<float> &iImageInput, image<float> &oImageOutput)
{
    int lImageWidth = iImageInput.column_size();
    int lImageHeight = iImageInput.row_size();

    if(lImageWidth != oImageOutput.column_size() || lImageHeight != oImageOutput.row_size())
    {
        std::cerr << "Image data format invalid - Input/Output" << std::endl;
        return;
    }

    std::vector< std::future<bool> > lResults;

    auto partial_filter = [this, &iImageInput, &oImageOutput](int iStartLine, int iEndLine)
    {
        partial_convolution_filter(iImageInput, oImageOutput, iStartLine, iEndLine);
        return true;
            
    };

    int lPacketSize = lImageHeight / NumberOfPackets;
    int iStartLine = 0, iEndLine = 0, i=0;
    int diff = lImageHeight-lPacketSize*NumberOfPackets;

    for (i=0 ; i<diff ; i++)
    {
        iEndLine += lPacketSize+1;
        lResults.emplace_back(mThreadPool->enqueue(partial_filter, iStartLine, iEndLine-1));
        iStartLine = iEndLine;
    }
    for(i = diff; i < NumberOfPackets; i++)
    {
        iEndLine += lPacketSize;
        lResults.emplace_back(mThreadPool->enqueue(partial_filter, iStartLine, iEndLine-1));
        iStartLine =  iEndLine;
    }
    for(auto && lResult: lResults) {lResult.wait();}
}

void Filter::applyMedianFilter(const image<float> &iImageInput, image<float> &oImageOutput)
{
    int lImageWidth = iImageInput.column_size();
    int lImageHeight = iImageInput.row_size();

    if(lImageWidth != oImageOutput.column_size() || lImageHeight != oImageOutput.row_size())
    {
        std::cerr << "Image data format invalid - Input/Output" << std::endl;
        return;
    }

    std::vector< std::future<bool> > lResults;

    auto partial_filter = [this, &iImageInput, &oImageOutput](int iStartLine, int iEndLine) 
    {
        partial_median_filter(iImageInput, oImageOutput, iStartLine, iEndLine); 
        return true;
    };

    int lPacketSize = lImageHeight / NumberOfPackets;
    int iStartLine = 0, iEndLine = 0, i=0;
    int diff = lImageHeight-lPacketSize*NumberOfPackets;

    for (i=0 ; i<diff ; i++)
    {
        iEndLine += lPacketSize+1;
        lResults.emplace_back(mThreadPool->enqueue(partial_filter, iStartLine, iEndLine-1));
        iStartLine = iEndLine;
    }
    for(i = diff; i < NumberOfPackets; i++)
    {
        iEndLine += lPacketSize;
        lResults.emplace_back(mThreadPool->enqueue(partial_filter, iStartLine, iEndLine-1));
        iStartLine =  iEndLine;
    }
    for(auto && lResult: lResults) {lResult.wait();}
}

void Filter::applyFilter(const image<float> &iImageInput, image<float> &oImageOutput)
{
    switch (filter_type)
    {
        case no_filter:
            NoFilter(iImageInput, oImageOutput);
            break;
        case convolution:
            applyConvolutionFilter(iImageInput, oImageOutput);
            break;
        case median:
            applyMedianFilter(iImageInput, oImageOutput);
            break;
        default:
            std::cout << "Warning wrong value for filter parameters, no gain apply" << std::endl;
            NoFilter(iImageInput, oImageOutput);
            break;
    }
}
    
//Convolution filter private functions
void Filter::partial_convolution_filter (const image<float> &ImageInput, image<float> &ImageOutput, int StartLine, int EndLine)
{
    int i=0, j=0;
    if (StartLine >= ImageInput.row_size()) {return;}
    if (EndLine > ImageInput.row_size()-1) {EndLine=ImageInput.row_size()-1;}
    if(StartLine == 0)
    {
        ImageOutput.data[0][0] = getWeightedValueOn2x2Sector(ImageInput, 0, 0);
        for (j=1 ; j<ImageInput.column_size()-1 ; j++)
        {
            ImageOutput.data[0][j] = getWeightedValueOn2x3Sector(ImageInput, 0, j);
        }
        ImageOutput.data[0][j] = getWeightedValueOn2x2Sector(ImageInput, 0, j);
        StartLine++;
    }
    for (i=StartLine ; i<EndLine ; i++)
    {
        ImageOutput.data[i][0] = getWeightedValueOn3x2Sector(ImageInput, i, 0);
        for (j=1 ; j<ImageInput.column_size()-1 ; j++)
        {
            ImageOutput.data[i][j] = getWeightedValueOn3x3Sector(ImageInput, i, j);
        }
        ImageOutput.data[i][j] = getWeightedValueOn3x2Sector(ImageInput, i, j);
    }
    if (i == ImageInput.row_size()-1)
    {
        ImageOutput.data[i][0] = getWeightedValueOn2x2Sector(ImageInput, i, 0);
        for (j=1 ; j<ImageInput.column_size()-1 ; j++)
        {
            ImageOutput.data[i][j] = getWeightedValueOn2x3Sector(ImageInput, i, j);
        }
        ImageOutput.data[i][j] = getWeightedValueOn2x2Sector(ImageInput, i, j);
    }
    else
    {
        ImageOutput.data[i][0] = getWeightedValueOn3x2Sector(ImageInput, i, 0);
        for (j=1 ; j<ImageInput.column_size()-1 ; j++)
        {
            ImageOutput.data[i][j] = getWeightedValueOn3x3Sector(ImageInput, i, j);
        }
        ImageOutput.data[i][j] = getWeightedValueOn3x2Sector(ImageInput, i, j); 
    }
}

float Filter::getWeightedValueOn2x2Sector(const image<float> &iImage, int iX, int iY) const
{
    float lWeightedSum = 0.f;
    if (iX == 0)
    {
        float lnorm=0.f;
        if (iY == 0)
        {
            lWeightedSum += iImage.data[0][0] * mKernel.data[1][1];
            lWeightedSum += iImage.data[0][1] * mKernel.data[1][2];
            lWeightedSum += iImage.data[1][0] * mKernel.data[2][1];
            lWeightedSum += iImage.data[1][1] * mKernel.data[2][2];
            lnorm = mKernel.data[1][1]+mKernel.data[1][2]+mKernel.data[2][1]+mKernel.data[2][2];
            lWeightedSum /= lnorm;
        }
        else if (iY == iImage.column_size()-1)
        {
            lWeightedSum += iImage.data[0][iY-1] * mKernel.data[1][0];
            lWeightedSum += iImage.data[0][iY] * mKernel.data[1][1];
            lWeightedSum += iImage.data[1][iY-1] * mKernel.data[2][0];
            lWeightedSum += iImage.data[1][iY] * mKernel.data[2][1];
            lnorm = mKernel.data[1][0]+mKernel.data[1][1]+mKernel.data[2][0]+mKernel.data[2][1];
            lWeightedSum /= lnorm;
        }
        else
        {
            std::cout << "Warning wrong 2x2 position asked for iX = 0" << std::endl;
        }
    }
    else if(iX == iImage.row_size()-1)
    {
        float lnorm=0.f;
        if (iY == 0)
        {
            lWeightedSum += iImage.data[iX-1][0] * mKernel.data[0][1];
            lWeightedSum += iImage.data[iX-1][1] * mKernel.data[0][2];
            lWeightedSum += iImage.data[iX][0] * mKernel.data[1][1];
            lWeightedSum += iImage.data[iX][1] * mKernel.data[1][2];
            lnorm = mKernel.data[0][1]+mKernel.data[0][2]+mKernel.data[1][1]+mKernel.data[1][2];
            lWeightedSum /= lnorm;
        }
        else if (iY == iImage.column_size()-1)
        {
            lWeightedSum += iImage.data[iX-1][iY-1] * mKernel.data[0][0];
            lWeightedSum += iImage.data[iX-1][iY] * mKernel.data[0][1];
            lWeightedSum += iImage.data[iX][iY-1] * mKernel.data[1][0];
            lWeightedSum += iImage.data[iX][iY] * mKernel.data[1][1];
            lnorm = mKernel.data[0][0]+mKernel.data[0][1]+mKernel.data[1][0]+mKernel.data[1][1];
            lWeightedSum /= lnorm;
        }
        else
        {
            std::cout << "Warning wrong 2x2 position asked for iX = end line" << std::endl;
        }
    }
    else
    {
        std::cout << "Warning wrong 2x2 position asked" << std::endl;
    }
    return lWeightedSum; 
}

float Filter::getWeightedValueOn2x3Sector(const image<float> &iImage, int iX, int iY) const
{
    float lWeightedSum = 0.f;
    int j=0;

    if (iX == 0)
    {
        j=1;
        for (int i=0 ; i<2 ; i++)
        {
            lWeightedSum += iImage.data[i][iY-1] * mKernel.data[j][0];
            lWeightedSum += iImage.data[i][iY] * mKernel.data[j][1];
            lWeightedSum += iImage.data[i][iY+1] * mKernel.data[j++][2];
        }
        lWeightedSum /= normalisation2x3top;
    }
    else if (iX == iImage.row_size()-1)
    {
        iX--;
        j=0;
        for (int i=iX ; i<iX+2 ; i++)
        {
            lWeightedSum += iImage.data[i][iY-1] * mKernel.data[j][0];
            lWeightedSum += iImage.data[i][iY] * mKernel.data[j][1];
            lWeightedSum += iImage.data[i][iY+1] * mKernel.data[j++][2];
        }
        lWeightedSum /= normalisation2x3bottom; 
    }
    else
    {
        std::cout << "Warning wrong 2x3 position asked" << std::endl;
    }
    return lWeightedSum;   
}

float Filter::getWeightedValueOn3x2Sector(const image<float> &iImage, int iX, int iY) const
{
    float lWeightedSum = 0.f;
    int j=0;

    if (iY == 0)
    {
        j=1;
        for (int i=0 ; i<2 ; i++)
        {
            lWeightedSum += iImage.data[iX-1][i] * mKernel.data[0][j];
            lWeightedSum += iImage.data[iX][i] * mKernel.data[1][j];
            lWeightedSum += iImage.data[iX+1][i] * mKernel.data[2][j++];
        }
        lWeightedSum /= normalisation3x2left;
    }
    else if (iY == iImage.column_size()-1)
    {
        iY--;
        j=0;
        for (int i=iY ; i<iY+2 ; i++)
        {
            lWeightedSum += iImage.data[iX-1][i] * mKernel.data[0][j];
            lWeightedSum += iImage.data[iX][i] * mKernel.data[1][j];
            lWeightedSum += iImage.data[iX+1][i] * mKernel.data[2][j++];
        }
        lWeightedSum /= normalisation3x2right; 
    }
    else
    {
        std::cout << "Warning wrong 3x2 position asked" << std::endl;
    }
    return lWeightedSum;   
}

float Filter::getWeightedValueOn3x3Sector(const image<float> &iImage, int iX, int iY) const
{
    float lWeightedSum = 0.f;
    int j=0;

    for (int i=iX-1 ; i<iX+2 ; i++)
    {
        lWeightedSum += iImage.data[i][iY-1] * mKernel.data[j][0];
        lWeightedSum += iImage.data[i][iY] * mKernel.data[j][1];
        lWeightedSum += iImage.data[i][iY+1] * mKernel.data[j++][2];
    }
    return lWeightedSum;
}

//Median private functions
void Filter::partial_median_filter (const image<float> &ImageInput, image<float> &ImageOutput, int StartLine, int EndLine)
{
    int i=0, j=0;
    std::vector<float> vector3x3(9);
    std::vector<float> vector2x3(6);
    std::vector<float> vector3x2(6);
    std::vector<float> vector2x2(4);
    if (StartLine >= ImageInput.row_size()) {return;}
    if (EndLine > ImageInput.row_size()-1) {EndLine = ImageInput.row_size()-1;}
    if(StartLine == 0)
    {
        ImageOutput.data[0][0] = getMedianOn2x2Neighboors(ImageInput.data, 0, 0, vector2x2);
        for (j=1 ; j<ImageInput.column_size()-1 ; j++)
        {
             ImageOutput.data[0][j] = getMedianOn2x3Neighboors(ImageInput.data, 0, j-1, vector2x3);
        }
        ImageOutput.data[0][j] = getMedianOn2x2Neighboors(ImageInput.data, 0, j-1, vector2x2);
        StartLine++;
    }
    for(i = StartLine; i < EndLine; i ++)
    {
        ImageOutput.data[i][0] = getMedianOn3x2Neighboors(ImageInput.data, i-1, 0, vector3x2);
        for(j = 1; j < ImageInput.column_size() - 1; j ++)
        {
            ImageOutput.data[i][j] = getMedianOn3x3Neighboors(ImageInput.data, i - 1, j - 1, vector3x3);
        }
        ImageOutput.data[i][j] = getMedianOn3x2Neighboors(ImageInput.data, i-1, j-1, vector3x2);
    }
    if(i == ImageInput.row_size()-1)
    {
        ImageOutput.data[i][0] = getMedianOn2x2Neighboors(ImageInput.data, i-1, 0, vector2x2);
        for (j=1 ; j<ImageInput.column_size() - 1; j++)
        {
             ImageOutput.data[i][j] = getMedianOn2x3Neighboors(ImageInput.data, i-1, j-1, vector2x3);
        }
        ImageOutput.data[i][j] = getMedianOn2x2Neighboors(ImageInput.data, i-1, j-1, vector2x2);
    }
    else
    {
        ImageOutput.data[i][0] = getMedianOn3x2Neighboors(ImageInput.data, i-1, 0, vector3x2);
        for(j = 1; j < ImageInput.column_size() - 1; j ++)
        {
            ImageOutput.data[i][j] = getMedianOn3x3Neighboors(ImageInput.data, i - 1, j - 1, vector3x3);
        }
        ImageOutput.data[i][j] = getMedianOn3x2Neighboors(ImageInput.data, i-1, j-1, vector3x2); 
    }
}

float Filter::getMedianOn3x3Neighboors(float ** iImage, int iTop, int iLeft, std::vector<float> &lNeighboors) const
{
    //use the top left corner of the 3x3 sector
    int pos=0;
    for(int i = iTop; i <= iTop+2; i++)
    {
        lNeighboors[pos++]=iImage[i][iLeft];
        lNeighboors[pos++]=iImage[i][iLeft+1];
        lNeighboors[pos++]=iImage[i][iLeft+2];
    }

    const auto lMedianIt = lNeighboors.begin() + 4;
    std::nth_element(lNeighboors.begin(), lMedianIt , lNeighboors.end());
    return *lMedianIt;
}

float Filter::getMedianOn3x3Neighboors(image<float> iImage, int iTop, int iLeft, std::vector<float> &lNeighboors) const
{
    //use the top left corner of the 3x3 sector
    int pos=0;
    for(int i = iTop; i <= iTop+2; i++)
    {
        lNeighboors[pos++]=iImage.data[i][iLeft];
        lNeighboors[pos++]=iImage.data[i][iLeft+1];
        lNeighboors[pos++]=iImage.data[i][iLeft+2];
    }

    const auto lMedianIt = lNeighboors.begin() + 4;
    std::nth_element(lNeighboors.begin(), lMedianIt , lNeighboors.end());
    return *lMedianIt;   
}

float Filter::getMedianOn2x2Neighboors(float **iImage, int iTop, int iLeft, std::vector<float> &lNeighboors) const
{
    //use the top left corner of the 2x2 sector
    int pos=0;
    for(int i = iTop; i <= iTop+1; i++)
    {
        lNeighboors[pos++]=iImage[i][iLeft];
        lNeighboors[pos++]=iImage[i][iLeft+1];
    }

    const auto lMedianIt = lNeighboors.begin() + 1;
    std::nth_element(lNeighboors.begin(), lMedianIt , lNeighboors.end());
    return *lMedianIt;
}

float Filter::getMedianOn2x2Neighboors(image<float> iImage, int iTop, int iLeft, std::vector<float> &lNeighboors) const
{
    //use the top left corner of the 2x2 sector
    int pos=0;
    for(int i = iTop; i <= iTop+1; i++)
    {
        lNeighboors[pos++]=iImage.data[i][iLeft];
        lNeighboors[pos++]=iImage.data[i][iLeft+1];
    }

    const auto lMedianIt = lNeighboors.begin() + 1;
    std::nth_element(lNeighboors.begin(), lMedianIt , lNeighboors.end());
    return *lMedianIt;    
}

float Filter::getMedianOn2x3Neighboors(float **iImage, int iTop, int iLeft, std::vector<float> &lNeighboors) const
{
    //use the top left corner of the 2x3 sector
    int pos=0;
    for(int i = iTop; i <= iTop+1; i++)
    {
        lNeighboors[pos++]=iImage[i][iLeft];
        lNeighboors[pos++]=iImage[i][iLeft+1];
        lNeighboors[pos++]=iImage[i][iLeft+2];
    }

    const auto lMedianIt = lNeighboors.begin() + 2;
    std::nth_element(lNeighboors.begin(), lMedianIt , lNeighboors.end());
    return *lMedianIt;
}

float Filter::getMedianOn2x3Neighboors(image<float> iImage, int iTop, int iLeft, std::vector<float> &lNeighboors) const
{
    //use the top left corner of the 2x3 sector
    int pos=0;
    for(int i = iTop; i <= iTop+1; i++)
    {
        lNeighboors[pos++]=iImage.data[i][iLeft];
        lNeighboors[pos++]=iImage.data[i][iLeft+1];
        lNeighboors[pos++]=iImage.data[i][iLeft+2];
    }

    const auto lMedianIt = lNeighboors.begin() + 2;
    std::nth_element(lNeighboors.begin(), lMedianIt , lNeighboors.end());
    return *lMedianIt;
}

float Filter::getMedianOn3x2Neighboors(float **iImage, int iTop, int iLeft, std::vector<float> &lNeighboors) const
{
    //use the top left corner of the 2x3 sector
    int pos=0;
    for(int i = iTop; i <= iTop+2; i++)
    {
        lNeighboors[pos++]=iImage[i][iLeft];
        lNeighboors[pos++]=iImage[i][iLeft+1];
    }

    const auto lMedianIt = lNeighboors.begin() + 2;
    std::nth_element(lNeighboors.begin(), lMedianIt , lNeighboors.end());
    return *lMedianIt;
}

float Filter::getMedianOn3x2Neighboors(image<float> iImage, int iTop, int iLeft, std::vector<float> &lNeighboors) const
{
    //use the top left corner of the 2x3 sector
    int pos=0;
    for(int i = iTop; i <= iTop+2; i++)
    {
        lNeighboors[pos++]=iImage.data[i][iLeft];
        lNeighboors[pos++]=iImage.data[i][iLeft+1];
    }

    const auto lMedianIt = lNeighboors.begin() + 2;
    std::nth_element(lNeighboors.begin(), lMedianIt , lNeighboors.end());
    return *lMedianIt;
}

