#include<iostream>
#include<cmath>

#include "GrayScale.hpp"

GrayScale::GrayScale(float nLowerBound, float nUpperBound, int nGrayLevel, QuantificationType nType)
{
    edit_Bound(nLowerBound, nUpperBound);
    edit_GrayLevel(nGrayLevel);
    edit_QuantificationType(nType);
    edit_number_of_packet(1);
    Scale=1.f;
    Slope=0.f;
}

void GrayScale::print()
{
    std::cout << "GrayScale object:" << std::endl;
    std::cout << "Quantification type: ";
    switch (Type)
    {
        case linear:
            std::cout << "Linear";
            break;
        case logaritmic:
            std::cout << "Logaritmic";
            break;
        case exponential:
            std::cout << "Exponential";
            break;
        case sigmoid:
            std::cout << "Sigmoïd";
            break;
        default:
            std::cout << "Unknow";
            break;
    }
    std::cout << std::endl;
    std::cout << "Number of gay level: " << GrayLevel << std::endl;
    std::cout << "Minimum pixel value: " << MinPixelValue << std::endl;
    std::cout << "Maximum pixel value: " << MaxPixelValue << std::endl;
    std::cout << "Lower bound: " << LowerBound << std::endl;
    std::cout << "Upper bound: " << UpperBound << std::endl;
}

void GrayScale::edit_Bound(float nLowerBound, float nUpperBound)
{
    if (nUpperBound <= nLowerBound)
    {
        nUpperBound = nLowerBound + 1.f;
        std::cout << "Warning, upper bound of LUT is inferior to lower bound, we take lower bound +1" << std::endl;
    }
    LowerBound = nLowerBound;
    UpperBound = nUpperBound;
}

void GrayScale::edit_GrayLevel(int nGrayLevel)
{
    if (nGrayLevel < 1)
    {
        std::cout << "Warning GrayLevel must be coded at least on 1 bit" << std::endl;
        nGrayLevel=1;
    }
    else if (nGrayLevel >8)
    {
        std::cout << "Warning GrayLevel can't be coded on more than 8 bits" << std::endl;
        nGrayLevel=8;
    }
    GrayLevel=nGrayLevel;
}

void GrayScale::edit_QuantificationType(QuantificationType nType)
{
    if ((nType < 0) || (nType >3))
    {
        std::cout << "Warning QuantificationType is 0, 1, 2 or 3" << std::endl;
        nType = linear;
    }
    Type = nType;
}

void GrayScale::setUp(std::shared_ptr<ThreadPool> iThreadPool)
{
    mThreadPool = iThreadPool;
}

void GrayScale::initFilter()
{
    switch (Type)
    {
        case linear:
            Scale=1.f;
            for (int i=0 ; i<GrayLevel ; i++) {Scale*=2.f;}
            Slope=(Scale-1.f)/(UpperBound-LowerBound);
            Scale=float(MaxPixelValue-MinPixelValue)/(Scale-1.f);
            break;
        case logaritmic: //bijective function of exponential
            OutMax=Exponential_factor;
            OutMin=0;
            Scale=1.f;
            for (int i=0 ; i<GrayLevel ; i++) {Scale*=2.f;}
            Slope=(Scale-1.f)/(OutMax-OutMin);
            Scale=float(MaxPixelValue-MinPixelValue)/(Scale-1.f);
            InternalSlope = (std::exp(Exponential_factor)-1.f)/(UpperBound-LowerBound);
            // internal slope to change input (from lower bound to upper bound) to [1,e(exponential_factor)]
            break;
        case exponential:
            OutMax=std::exp(Exponential_factor);
            OutMin=1.f;
            Scale=1.f;
            for (int i=0 ; i<GrayLevel ; i++) {Scale*=2.f;}
            Slope=(Scale-1.f)/(OutMax-OutMin);
            Scale=float(MaxPixelValue-MinPixelValue)/(Scale-1.f);
            InternalSlope = Exponential_factor/(UpperBound-LowerBound);
            // internal slope to change input (from lower bound to upper bound) to [0,Exponential_factor]
            break;
        case sigmoid:
            Scale=1.f;
            for (int i=0 ; i<GrayLevel ; i++) {Scale*=2.f;}
            Slope=Scale-1.f;
            Scale=float(MaxPixelValue-MinPixelValue)/(Scale-1.f);
            InternalSlope=11.08252f/(UpperBound-LowerBound);
            break; 
        default:
            std::cout << "Warning, quantification type not implemented..." << std::endl;
    }
}

int GrayScale::LinearQuantification(float value)
{
    int out=0;
    if (value <= LowerBound) {return MinPixelValue;}
    else if (value >= UpperBound) {return MaxPixelValue;}
    else 
    {
        out=int(Slope*(value-LowerBound));
        return ((int)(Scale*out));
    }
}

int GrayScale::LogarithmicQuantification(float value)
{
    int out=0;
    if (value <= LowerBound) {return MinPixelValue;}
    else if (value >= UpperBound) {return MaxPixelValue;}
    else
    {
        value = (value-LowerBound)*InternalSlope+1.f;
        out=int(Slope*(std::log(value)-OutMin));
        return ((int)(Scale*out));
    }
}

int GrayScale::ExponentialQuantification(float value)
{
    int out=0;
    if (value <= LowerBound) {return MinPixelValue;}
    else if (value >= UpperBound) {return MaxPixelValue;}
    else
    {
        value = (value-LowerBound)*InternalSlope;
        out=int(Slope*(std::exp(value)-OutMin));
        return ((int)(Scale*out));
    }
}

int GrayScale::SigmoidQuantification(float value)
{
    int out=0;
    if (value <= LowerBound) {return MinPixelValue;}
    else if (value >= UpperBound) {return MaxPixelValue;}
    else
    {
        value=(value-LowerBound)*InternalSlope-5.54126f; 
        out=int(Slope*(1.f/(1.f+std::exp(-value))));
        return ((int)(Scale*out));
    }
}

int GrayScale::get_output(float value)
{
    int out=0;
    switch (Type)
    {
        case linear:
            out = LinearQuantification(value);
            break;
        case logaritmic:
            out = LogarithmicQuantification(value);
            break;
        case exponential:
            out = ExponentialQuantification(value);
            break;
        case sigmoid:
            out = SigmoidQuantification(value);
            break;
        default:
            out = int(value);
            break;
    }
    return out;
}

void GrayScale::applyFilter(const image<float> &iImageInput, image<int> &oImageOutput)
{
  int lImageWidth = iImageInput.column_size();
  int lImageHeight = iImageInput.row_size();

  if(lImageWidth != oImageOutput.column_size() || lImageHeight != oImageOutput.row_size())
  {
    std::cerr << "Image data format invalid - Input/Output" << std::endl;
    return;
  }


  std::vector< std::future<bool> > lResults;
  // lambda used to process partial data packet
  // TODO: may required optimisation on data input / output
  auto partialLinearQuantification = [&](int iStartLine, int iEndLine)
  {
    for(int i = iStartLine; i < iEndLine; i ++)
    {
      for(int j = 0; j < iImageInput.column_size(); j ++)
      {
        switch (Type)
        {
            case linear:
                oImageOutput.data[i][j]=LinearQuantification(iImageInput.data[i][j]);
                break;
            case logaritmic:
                oImageOutput.data[i][j]=LogarithmicQuantification(iImageInput.data[i][j]);
                break;
            case exponential:
                oImageOutput.data[i][j]=ExponentialQuantification(iImageInput.data[i][j]);
                break;
            case sigmoid:
                oImageOutput.data[i][j]=SigmoidQuantification(iImageInput.data[i][j]);
                break;
            default:
                oImageOutput.data[i][j]=int(iImageInput.data[i][j]);
                break;
        }
      }
    }

      return true;
  };

  int lPacketSize = lImageHeight / NumberOfPackets;
  int iStartLine = 0, iEndLine = 0, i=0;
  int diff = lImageHeight-lPacketSize*NumberOfPackets;
  for (i=0 ; i<diff ; i++)
  {
    iEndLine += lPacketSize+1;
    lResults.emplace_back(
      mThreadPool->enqueue(partialLinearQuantification, iStartLine, iEndLine)
    );
    iStartLine =  iEndLine;
  }
  for(int i = diff; i < NumberOfPackets; i++)
  {
    iEndLine += lPacketSize;
    lResults.emplace_back(
      mThreadPool->enqueue(partialLinearQuantification, iStartLine, iEndLine)
    );
    iStartLine =  iEndLine;
  }

  for(auto && lResult: lResults)
  {
    lResult.wait();
  }

}
