#ifndef MATRIX_HPP
#define MATRIX_HPP

/*
 * contributors:    - Clement Le Couedic
 *                  - Jerome Dubois
 * version: 1.2
 * date: 03/2020
 *
 *
 */

#include<iostream>
#include<cstring> //for memcpy
#include"pointer_management.hpp"

//Forward declaration
template <class T> class image;
template <class T> std::ostream& operator<<(std::ostream &oOS, const image<T>& nimage);

//class declaration
template <class T>
class image
{
  private:
    int Nrow;
    int Ncolumn;
    bool verbose;

  public:
    T **data;

    image(int nNrow=1, int nNcolumn=1); //constructor
    image(const image<T> &other); //copy constructor
    image& operator=(const image<T> &other); //copy assignement operator
    image(image<T> &&other); //move constructor
    image& operator=(image<T> &&other); //move assignement operator
    ~image();

    //opertor
    T operator() (int id_row, int id_column) const;
    T& operator() (int id_row, int id_column);
    friend std::ostream& operator << <> (std::ostream &oOS, const image<T>& nimage);

    //general function
    inline void set_verbose_on() {verbose = true;}
    inline void set_verbose_off() {verbose = false;}
    inline bool get_verbose() {return verbose;}
    void resize(int nNrow, int nNcolumn);
    int row_size() const {return Nrow;}
    unsigned int unsigned_row_size() const {return (unsigned int)(Nrow);}
    int column_size() const {return Ncolumn;}
    unsigned int unsigned_column_size() const {return (unsigned int)(Ncolumn);}

    void reset(T value=0); //reset all the matrix to value
    T sum(); //return sum of all elements of the matrix
    T sumd(); //return sum of all elements, but calculus in double precision

    void write(std::string);
    void load(std::string, int nNrow, int nNcolumn);
};

//Constructor
template <class T>
image<T>::image(int nNrow, int nNcolumn)
{
    Nrow=nNrow;
    Ncolumn=nNcolumn;
    create_matrix(data, Nrow, Ncolumn);
    set_verbose_off();
}

//Copy constructor
template <class T>
image<T>::image(const image<T> &other)
{
    Nrow = other.Nrow;
    Ncolumn = other.Ncolumn;
    verbose = other.verbose;

    //vector
    create_matrix(data, Nrow, Ncolumn);
    for (int i=0 ; i<Nrow ; i++) {std::memcpy(data[i], other.data[i], Ncolumn*sizeof(T));};
}

//Copy assignement operator
template <class T>
image<T>& image<T>::operator=(const image<T> &other)
{
    if (this!=&other)
    {
        if ((Nrow != other.Nrow) || (Ncolumn!=other.Ncolumn))
        {
            resize_matrix(data, Nrow, Ncolumn, other.Nrow, other.Ncolumn);
            Nrow = other.Nrow;
            Ncolumn = other.Ncolumn;
            verbose = other.verbose;
        }
        for (int i=0 ; i<Nrow ; i++) {std::memcpy(data[i], other.data[i], Ncolumn*sizeof(T));};
    }
    return *this;
}

//Move constructor
template <class T>
image<T>::image(image<T> &&other)
{
    Nrow = other.Nrow;
    Ncolumn = other.Ncolumn;
    verbose = other.verbose;
    data = other.data;

    //release other ressources
    other.Nrow = 0;
    other.Ncolumn = 0;
    other.verbose = false;
    other.data = nullptr;
}

//Move assignement operator
template <class T>
image<T>& image<T>::operator=(image<T> &&other)
{
    if (this != &other)
    {
        delete_matrix(data, Nrow);
        Nrow = other.Nrow;
        Ncolumn = other.Ncolumn;
        verbose = other.verbose;
        data = other.data;

        //release other image ressources
        other.Nrow = 0;
        other.Ncolumn = 0;
        other.verbose = false;
        other.data = nullptr;
    }
    return *this;
}

//Destructor
template <class T>
image<T>::~image()
{
    delete_matrix(data, Nrow);
    Nrow = 0;
    Ncolumn = 0;
    verbose = false;
}

//operators
template <class T>
T image<T>::operator() (int id_row, int id_column) const
{
    if (id_row < 0)
    {
        if (verbose)
        {
            std::cout << "Warning, index of row out of bound, trying to access negative index: " << id_row << std::endl;
        }
        id_row = 0;
    }
    else if (id_row >= Nrow)
    {
        if (verbose)
        {
            std::cout << "Warning, index of row out of bound, trying to access element: " << id_row+1 << " (ID " << id_row << ") whereas number of element is: " << Nrow << std::endl;
        }
        id_row = Nrow-1;
    }
    if (id_column < 0)
    {
        if (verbose)
        {
            std::cout << "Warning, index of column out of bound, trying to access negative index: " << id_column << std::endl;
        }
        id_column = 0;
    }
    else if (id_column >= Ncolumn)
    {
        if (verbose)
        {
            std::cout << "Warning, index of column out of bound, trying to access element: " << id_column+1 << " (ID " << id_column << ") whereas number of element is: " << Ncolumn << std::endl;
        }
        id_column = Ncolumn-1;
    }
    return data[id_row][id_column];
}

template <class T>
T& image<T>::operator() (int id_row, int id_column)
{
    if (id_row < 0)
    {
        if (verbose)
        {
            std::cout << "Warning, index of row out of bound, trying to access negative index: " << id_row << std::endl;
        }
        id_row = 0;
    }
    else if (id_row >= Nrow)
    {
        if (verbose)
        {
            std::cout << "Warning, index of row out of bound, trying to access element: " << id_row+1 << " (ID " << id_row << ") whereas number of element is: " << Nrow << std::endl;
        }
        id_row = Nrow-1;
    }
    if (id_column < 0)
    {
        if (verbose)
        {
            std::cout << "Warning, index of column out of bound, trying to access negative index: " << id_column << std::endl;
        }
        id_column = 0;
    }
    else if (id_column >= Ncolumn)
    {
        if (verbose)
        {
            std::cout << "Warning, index of column out of bound, trying to access element: " << id_column+1 << " (ID " << id_column << ") whereas number of element is: " << Ncolumn << std::endl;
        }
        id_column = Ncolumn-1;
    }
    return data[id_row][id_column];
}

template<class T>
std::ostream& operator<<(std::ostream &oOS, const image<T>& nimage)
{
    oOS << "image Nrow =" << nimage.Nrow << ", Ncolumn =" << nimage.Ncolumn << ";" <<std::endl;
    for (int i=0 ; i<nimage.Nrow ; i++)
    {
        for (int j=0 ; j<nimage.Ncolumn ; j++) {oOS << nimage.data[i][j] << " ";}
        oOS << std::endl;
    }
    oOS << std::endl;
    return oOS;
}

//general functions
template <class T>
void image<T>::reset(T value)
{
    for (int i=0 ; i<Nrow ; i++)
    {
        for (int j=0 ; j<Ncolumn ; j++) {data[i][j]=T(value);}
    }
}

template <class T>
void image<T>::resize(int nNrow, int nNcolumn)
{
    if ((Nrow != nNrow) || (Ncolumn != nNcolumn))
    {
        resize_matrix(data, Nrow, Ncolumn, nNrow, nNcolumn);
        Nrow=nNrow;
        Ncolumn=nNcolumn;
    }
}

template<class T>
T image<T>::sum()
{
    T sum = T(0);
    for (int i=0 ; i<Nrow ; i++)
    {
        for (int j=0 ; j<Ncolumn ; j++) {sum+=data[i][j];}
    }
    return sum;
}

template<class T>
T image<T>::sumd()
{
    double sum = 0.0;
    for (int i=0 ; i<Nrow ; i++)
    {
        for (int j=0 ; j<Ncolumn ; j++) {sum+=double(data[i][j]);}
    }
    return T(sum);
}

template <class T>
void image<T>::write(std::string name)
{
    std::ofstream txtfile;
    txtfile.open(name);
    if (!txtfile)
    {
        std::cout << "Error opening file " << name << " to write line<T>" << std::endl;
        exit(1);
    }

    //txtfile << Nel << std::endl;
    for (int i=0 ; i<Nrow ; i++) 
    {
        for (int j=0 ; j<Ncolumn ; j++)
        {
            txtfile << data[i][j] << " ";
        }
        txtfile << std::endl;
    }
    txtfile.close();
}

template <class T>
void image<T>::load(std::string name, int nNrow, int nNcolumn)
{
    std::ifstream txtfile;
    txtfile.open(name);
    if (!txtfile)
    {
        std::cout << "Error opening file " << name << " to write line<T>" << std::endl;
        exit(1);
    }

    //txtfile << Nel << std::endl;
    
    Nrow = nNrow;
    Ncolumn = nNcolumn;
    for (int i=0 ; i<Nrow ; i++) 
    {
        for (int j=0 ; j<Ncolumn ; j++)
        {
            txtfile >> data[i][j];
        }
    }
    txtfile.close();
}

#endif
