#ifndef SCAN_CONV_FLOAT
#define SCAN_CONV_FLOAT

#include"pointer_management.hpp"
#include <memory>
#include"ThreadPool.h"
#include"image.hpp"
//#include"../json/json.hpp"

/*-----
Contributor: Jérôme Dubois
Date: 07/2019
Version: 1.0

Description:
C++ class create to make scan conversion (convert ultrasound data in random space to image in cartesian
space). The scan conversion can be made with bi_linear interpolation or weight interpolation. The scan
conversion consider 3 possible mechanism:
- pure rotation (us data in cylindrical space)
- linear displacement (us data in cylindrical space, a translated char make a pseudo circular movement)
- lounger (us data come from transducer on a lounger, not a circular movement)

Usage:

using ultrasound card parameters:
- scan_conv sc(int Nline, int Npoint, int Nx, int Ny)
- sc.edit_sampling_us(float sampling_rate, int decimation, float speed, int Npoint)
- sc.edit_delay_us(float delay)
- sc.edit_mechanism(Mechanism mechanism, float h0, float dx, float rb, float li, float ti)
- sc.edit_scanning(float sector, int Nline, ScanningOption scanning_option)
- sc.edit_depth(float R0, float Rf)
- sc.kernel_calculation()
- std::shared_ptr<ThreadPool> sc_Pool = std::make_shared<ThreadPool>(8)
- sc.setUp(sc_Pool);
- sc.image_scan_conversion(float **image)

using mm parameters:
- scan_conv sc(int Nline, int Npoint, int Nx, int Ny)
- sc.edit_sampling_mm(float sampling_mm, int nNpoint)
- sc.edit_delay_mm(float delay)
- sc.edit_mechanism(int mechanism, float h0, float dx, float rb, float li, float ti)
- sc.edit_scanning(float sector, int Nline, int scanning_option)
- sc.edit_depth(float &R0, float &Rf)
- sc.kernel_calculation()
- std::shared_ptr<ThreadPool> sc_Pool = std::make_shared<ThreadPool>(8)
- sc.setUp(sc_Pool);
- sc.image_scan_conversion(float **image)


for pure rotation:
- edit_mechanism(0, float h0)
for linear displacement (here the transducer in fixe on the arm not on the chart, it make a pure rotation)
- edit_mechanism(1, float h0, float dx)
for lounger
- edit_mechanism(2, float h0, float dx, float rb, float li, float ti)

if scanning_option=0 delta theta is constant in the image
if scanning_option=1 delta x of the chart is constant
image_scan_conversion is multi-threaded, one can use image_scan_conversion_raw wich is not multi-threaded

size of us image can be change with 
- sc.edit_us_image_parameters(int Nline, int Npoint)

size of final image can be change with
- sc.edit_image_size(int Nx, int Ny)

if a parameter is changed, kernel_calculation() must be resend before a scan conversion, but need to be done only once before making any number of scan_conversion without changing parameters.

-----*/

enum Mechanism
{
    pure_rotation = 0,
    translation_rotation,
    lounger
};

enum InterpolationType
{
    bilinear = 0,
    distance_weight
};

enum ScanningOption
{
    constant_theta = 0,
    constant_elevation
};

class scan_conv
{
    private:
        //thread parameters
        int NumberOfPackets;
        std::shared_ptr<ThreadPool> mThreadPool;
        
        //scan conversion
        int error; //Not implemented yet
        InterpolationType interpolation_type; //bilinear or distance/weight interpolation
        bool square;

        //image parameter
        int Nx; //nomber of pixel in the image along x
        int Ny; //nomber of pixel in the image along y
        float R0; //start depth of the scan conversion
        float Rf; //end depth of the scan convers
        float Rmin; //minimum depth of the scan conversion, R0 >= Rmin
        float Rmax; //maximum depth of the scan conversion, Rf >= Rmax

        //mechanism parameters
        Mechanism mechanism; //mechanism consider by the scan conversion
        ScanningOption scanning_option; //defined geometry of scanning line (angle between each line), 0 constant theta, 1 constant elevation
        float h0; //height of the central point compare to cart in mm
        //float hp; //height of the nose of the probe compare to h0 in mm not used actually
        float sector; //sector of the image displayed in °
        float sector_rd; //sector of the image displayed in rad
        float total_sector; //sector swept by the transducer in °
        float total_sector_rd; //sector swept by the transducer in rad
        float dx; //total translation distance in mm
        float rb; //distance between pivot (central point) and transducer in mm
        float li; //lateral distance between tranducer and lounger arm in mm
        float ti; //angle between transducer and loungner arm in °
        float ti_rd; //angle between transducer and lounger arm in rad
        float rtang; //radium of the tangent circle describe by lines for mechanism 2 case in mm
        float x0; //laterale shift of cross point compare to center of the nose in mm
        
        //ultrasound parameters
        int Npoint; //nomber of point per lines
        int Nline; //nomber of line per images
        int decimation; //decimation made by the ultrasound board
        float sampling_rate; //sampling rate of the ultrasound board in MHz
        float true_sampling_rate; // equal to sampling_rate / decimation
        float speed; //speed of sound in the media in m/s
        float dr; //sampling distance in mm
        float delay_us; //delay between pulse and beginning of sampling in us
        float delay_mm; //distance between pulse and beginning of sampling in mm


        //to see
        int N_point_to_change; //number of pixel to calculate in the scan_conversion area
        int N_out; //number of pixel outside the scan_conversion area

        float *x_vector; //x_vector of the image, Nx points (regulary space)
        float *y_vector; //y_vector of the image, Ny points (regulary space)

        int **indicial_x_y_tmp; //temporary matrix to store pixel in the scan conversion area, too large, 2*(Nx*Ny)
        //int **indicial_x_y; //matrix 2*N_point_to_change to calculate image
        int **indicial_x_y_out_tmp; //temporary matrix to store pixel out of the scan conversion area too large, 2*(Nx*Ny)
        //int **indicial_x_y_out; //matrix 2*N_out to change image backgroune
        int **indicial_r_theta; //matrix 4*N_point_to_change indicial of the 4 points in the r theta plan that are used in the calculus of the Nth pixel, pixel listed in indicial_x_y
        float **weight; //matrix 4*N_point_to_change, weight of the 4 points listed in indicial_r_theta for the calcul of the pixel listed in indicial_x_y
        float **r_vector;
        float **theta_vector;
        float **r_matrix;
        float **theta_matrix;
        float *xi; //abscissa of transducer shot points
        float *yi; //ordinate of transducer shot points
        float *ri; //distance between transducer and cross point of the line shot point
        float *thetai; //angle of the transducer shot lines

    public:
        float **Image; //output image
        
        scan_conv(int nNline=64, int nNpoint=500, int nNx=512, int nNy=512); //by defaut we consider pure rotation mechanism
        ~scan_conv();
        void print();
        void print_position();
        void set_image_square() {square = true;}
        void set_image_arbitrary() {square = false;}
        //resize(int nNline, int nNpoint, int nNx, int nNy);
        
        //Thread
        void edit_number_of_packet(int np) {NumberOfPackets = np;}
        void setUp(std::shared_ptr<ThreadPool> iThreadPool);

        void edit_Npoint(int nNpoint);
        void edit_Nline(int nNline);
        void edit_speed(float nspeed) {speed = nspeed;}
        void edit_sampling_us(float nsampling_rate, int ndecimation, float nspeed, int nNpoint);
        void edit_sampling_mm(float sampling_mm, int nNpoint);
        void edit_mechanism(Mechanism nmechanism = pure_rotation, float nrb = 0.f, float ndx = 0.f, float nh0 = 0.f, float nli = 0.f, float nti = 0.f);
        void edit_scanning(float nsector, int nNline, ScanningOption nscanning_option);
        void edit_delay_us(float ndelay);
        void edit_delay_mm(float ndelay);
        void calcul_maximum_depth();
        void calcul_transducer_position();
        float vector_min(float *in); //give the minimum of the vector, Nline elements
        float vector_max(float *in); //give the maximum of the vector, Nline elements
        void edit_depth(float nR0, float nRf);
        void get_depth(float &lR0, float &lRf) {lR0=R0; lRf=Rf;}
        void get_maximum_depth(float &lRmin, float &lRmax);
        void edit_image_size(int nNx, int nNy);
        void edit_us_image_parameters(int nNline, int nNpoint);

        void set_bilinear_interpolation() {interpolation_type=bilinear;}
        void set_weight_interpolation() {interpolation_type=distance_weight;}

        void fill_xy_vector();
        void fill_rtheta_vectors();
        void fill_rtheta_matrices();
        void indicial_matrix_calculation();
        void weight_matrix_calculation();
        void kernel_calculation();

        void image_scan_conversion_raw(float **image_r_theta);
        void partial_scan_conversion(float **image_r_theta, int start_point, int end_point);
        void image_scan_conversion(float **image_r_theta);
        void change_image_background(float value);
        void change_image_value(float value);

        void image_scan_conversion_raw(const image<float> &input, image<float> &output);
        void partial_scan_conversion(const image<float> &input, image<float> &output, int start_point, int end_point);
        void image_scan_conversion(const image<float> &input, image<float> &output);
        void change_image_background(image<float> &input, float background);
        void change_image_background(const image<float> &input, image<float> &output, float background);
        void change_image_value(image<float> &input, float value);
        void change_image_value(const image<float> &input, image<float> &output, float value);

        void save_image(std::string name);
        void save_rmatrix(std::string name);
        void save_tmatrix(std::string name);
    
        //inline functions
        inline int getNx() {return Nx;}
        inline int getNy() {return Ny;}
        inline int getNline() {return Nline;}
        inline int getNpoint() {return Npoint;}
        inline float** getImage(){return Image;}

        inline int get_number_of_packet() {return NumberOfPackets;}
        inline float get_speed() {return speed;}
        inline Mechanism get_mechanism() {return mechanism;}
        inline float get_h0() {return h0;}
        inline float get_dx() {return dx;}
        inline float get_rb() {return rb;}
        inline float get_li() {return li;}
        inline float get_ti() {return ti;}
        inline float get_sector() {return sector;}
        inline ScanningOption get_scanning_option() {return scanning_option;}
        inline bool is_square() {return square;}
        inline float get_R0() {return R0;}
        inline float get_Rf() {return Rf;}
        inline float get_sampling_rate() {return sampling_rate;}
        inline int get_decimation() {return decimation;}
        inline float get_delay_us() {return delay_us;}
        inline float get_delay_mm() {return delay_mm;}
};

#endif
