#include<gtest/gtest.h>

#include<random> // need c++11
#include<cmath> // for cos

#include"line.hpp"
#include"SigProc1D.hpp"

/*********
author: Jérôme Dubois
date: March 2020



********/

// Actually, test only made on line<float> based functions
// TODO make test for 2 other or delete them

double relative_error(double x, double y);

TEST( Sigproc1D_Constructor, size)
{
    //default size (1)
    SigProc1D sp1;
    EXPECT_EQ(sp1.get_internal_vector_size(), 1);
    line<float> lgain;
    lgain = sp1.get_gain();
    EXPECT_EQ(lgain.size(), 1);
    //random size
    int npoint = 256;
    SigProc1D sp2(npoint);
    EXPECT_EQ(sp2.get_internal_vector_size(), npoint);
    lgain = sp2.get_gain();
    EXPECT_EQ(lgain.size(), 1);
}

TEST( Sigproc1D_Resize, size)
{
    // greater size
    int npoint = 4567;
    SigProc1D sp1(12);
    line<float> lgain;
    line<float> nvec(npoint);
    sp1.edit_raw_data_vector(nvec);
    lgain = sp1.get_gain();
    EXPECT_EQ(sp1.get_internal_vector_size(), npoint);
    EXPECT_EQ(lgain.size(), 1);
    // smaller size
    npoint = 159;
    nvec.resize(npoint);
    sp1.edit_raw_data_vector(nvec);
    lgain = sp1.get_gain();
    EXPECT_EQ(sp1.get_internal_vector_size(), npoint);
    EXPECT_EQ(lgain.size(), 1);
}

TEST( Sigproc1D_basic_functions, shift_data )
{
    int npoint = 42;
    SigProc1D sp1;
    line<float> input_line(npoint), output_line(1);
    float shift = 0.354f;
    //juste a rapid test because it use scalar + from line.hpp already validated
    for (int i=0 ; i<npoint ; i++)
    {
        input_line(i) = 56.214f*float(i);
    }
    sp1.shift_data(input_line, output_line, shift);
    EXPECT_EQ(npoint, output_line.size());
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(shift+input_line(i), output_line(i)) << "output_line[" << i <<"] not equal to shifted value";
    }

    // internal function
    sp1.edit_raw_data_vector(input_line);
    sp1.shift_data(shift);
    line<float> internal_vector = sp1.get_filtered_data();
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(internal_vector(i), output_line(i)) << "error internal function at loop " << i;
    }

    // external pointers
    float *input_p=nullptr, *output_p=nullptr;
    create_vector(input_p, npoint);
    create_vector(output_p, npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        input_p[i] = input_line(i);
    }
    sp1.shift_data(input_p, output_p, npoint, shift);
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( output_p[i], output_line(i) ) << "error external pointers at loop " << i;
    }
    delete_vector(input_p);
    delete_vector(output_p);
}

TEST( Sigproc1D_basic_functions, center_data )
{
    int npoint = 42;
    SigProc1D sp1;
    line<float> input_line(npoint), output_line(1), center_line(npoint);;
    for (int i=0 ; i<npoint-1 ; i+=2)
    {
        input_line(i) = 5.f;
        input_line(i+1) = 10.f;
        center_line(i) = -2.5f;
        center_line(i+1) = 2.5f;
    }
    //juste a rapid test because it use mean() function and scalar  from line.hpp already validated
    sp1.center_data(input_line, output_line);
    EXPECT_EQ(npoint, output_line.size());
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_EQ(center_line(i), output_line(i)) << "output_line[" << i <<"] not equal to expected centered value";
    }

    // internal function
    sp1.edit_raw_data_vector(input_line);
    sp1.center_data();
    line<float> internal_vector = sp1.get_filtered_data();
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(internal_vector(i), center_line(i)) << "error internal function at loop " << i;
    }

    // external pointers
    float *input_p=nullptr, *output_p=nullptr;
    create_vector(input_p, npoint);
    create_vector(output_p, npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        input_p[i] = input_line(i);
    }
    sp1.center_data(input_p, output_p, npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( output_p[i], center_line(i) ) << "error external pointers at loop " << i;
    }
    delete_vector(input_p);
    delete_vector(output_p);
}

TEST( Sigproc1D_basic_functions, absolute_value )
{
    // prepare random generetor c++11
    std::random_device rd;
    std::mt19937 e2(rd());
    std::uniform_real_distribution<> dist(-10, 0);

    int npoint = 69;
    SigProc1D sp1;
    line<float> input_line(npoint), output_line(1);
    for (int i=0 ; i<npoint ; i++)
    {
        input_line(i) = dist(e2);
    }
    line<float> zero_vec(npoint), abs_vec(npoint);
    abs_vec = zero_vec-input_line; // input line is always < 0 so abolute value is - input_line
    sp1.absolute_value(input_line, output_line);
    EXPECT_EQ(npoint, output_line.size());
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_EQ(abs_vec(i), output_line(i)) << "output_line[" << i <<"] not equal to expected absolute value";
    }

    // internal function
    sp1.edit_raw_data_vector(input_line);
    sp1.absolute_value();
    line<float> internal_vector = sp1.get_filtered_data();
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(internal_vector(i), output_line(i)) << "error internal function at loop " << i;
    }

    // external pointers
    float *input_p=nullptr, *output_p=nullptr;
    create_vector(input_p, npoint);
    create_vector(output_p, npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        input_p[i] = input_line(i);
    }
    sp1.absolute_value(input_p, output_p, npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( output_p[i], output_line(i) ) << "error external pointers at loop " << i;
    }
    delete_vector(input_p);
    delete_vector(output_p);
}

TEST( Sigproc1D_basic_function, normalize )
{
    int npoint = 1803;
    SigProc1D sp1;
    // x for linear function
    line<float> x_vec(npoint);
    for (int i=0 ; i<npoint ; i++) {x_vec(i) = float(i - npoint/100);}

    // increasing function, max at the end
    line<float> y_vec(npoint), y_vec_norm(npoint), result;
    float d1 = 12.65f, dn = 1.f / x_vec(npoint-1);
    for (int i=0 ; i<npoint ; i++)
    {
        y_vec(i) = d1 * x_vec(i);
        y_vec_norm(i) = dn * x_vec(i);
    }
    sp1.normalize( y_vec, result);
    EXPECT_EQ( npoint, result.size() );
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(y_vec_norm(i), result(i)) << "output_line[" << i <<"] not equal to expected normalized value";
    }
    // increasing function, max at beggining
    for (int i=0 ; i<npoint ; i++) {x_vec(i) = float(i-npoint+15);}
    dn = -1.f / x_vec(0); // minus becaus x_vec is negative at beggining
    for (int i=0 ; i<npoint ; i++)
    {
        y_vec(i) = d1 * x_vec(i);
        y_vec_norm(i) = dn * x_vec(i);
    }
    sp1.normalize( y_vec, result);
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(y_vec_norm(i), result(i)) << "output_line[" << i <<"] not equal to expected normalized value";
    }

    // decreasing function, max at the end
    for (int i=0 ; i<npoint ; i++) {x_vec(i) = float(i-npoint/114);}
    d1 = -57.65f;
    dn = -1.f / x_vec(npoint-1);
    for (int i=0 ; i<npoint ; i++)
    {
        y_vec(i) = d1 * x_vec(i);
        y_vec_norm(i) = dn * x_vec(i);
    }
    sp1.normalize( y_vec, result);
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(y_vec_norm(i), result(i)) << "output_line[" << i <<"] not equal to expected normalized value";
    }
  
    // cosinus function
    for (int i=0 ; i<npoint ; i++) {x_vec(i) = float(i-npoint/4);}
    d1 = 3.14159f;
    for (int i=0 ; i<npoint ; i++)
    {
        y_vec(i) = d1 * std::cos(x_vec(i));
        y_vec_norm(i) =  std::cos(x_vec(i));
    }
    sp1.normalize( y_vec, result);
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(y_vec_norm(i), result(i)) << "output_line[" << i <<"] not equal to expected normalized value";
    }

    // internal function
    sp1.edit_raw_data_vector(y_vec);
    sp1.normalize();
    line<float> internal_vector = sp1.get_filtered_data();
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(internal_vector(i), y_vec_norm(i)) << "error internal function at loop " << i;
    }

    // external pointers
    float *input_p=nullptr, *output_p=nullptr;
    create_vector(input_p, npoint);
    create_vector(output_p, npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        input_p[i] = y_vec(i);
    }
    sp1.normalize(input_p, output_p, npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( output_p[i], y_vec_norm(i) ) << "error external pointers at loop " << i;
    }
    delete_vector(input_p);
    delete_vector(output_p);

}

TEST( Sigproc1D_basic_function, get_dB )
{
    float target = 6.02059991327962f;
    float value = 2.f, output = 0.f;
    SigProc1D sp1;
    sp1.get_dB(&value, &output, 1);
    EXPECT_FLOAT_EQ(target, output);

    // prepare random generetor c++11
    std::random_device rd;
    std::mt19937 e2(rd());
    std::uniform_real_distribution<> dist(0.1f, 10.f);

    int npoint = 581;
    line<float> input_line(npoint), rebuild_line(npoint), output_line;
    for (int i=0 ; i<npoint ; i++)
    {
        input_line(i) = dist(e2);
    }
    sp1.get_dB(input_line, output_line);
    EXPECT_EQ( npoint, output_line.size() );
    for (int i=0 ; i<npoint ; i++)
    {
        rebuild_line(i) = std::pow(10.f, output_line(i)/20.f);
    }
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(input_line(i), rebuild_line(i)) << "output_line[" << i <<"] not equal to expected normalized value";
    }

    // internal function
    sp1.edit_raw_data_vector(input_line);
    sp1.get_dB();
    line<float> internal_vector = sp1.get_filtered_data();
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(internal_vector(i), output_line(i)) << "error internal function at loop " << i;
    }

    // external pointers
    float *input_p=nullptr, *output_p=nullptr;
    create_vector(input_p, npoint);
    create_vector(output_p, npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        input_p[i] = input_line(i);
    }
    sp1.get_dB(input_p, output_p, npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( output_p[i], output_line(i) ) << "error external pointers at loop " << i;
    }
    delete_vector(input_p);
    delete_vector(output_p);
}

TEST( Sigproc1D_basic_function, apply_constant_gain )
{
    int npoint = 942;
    float value = 42.69f;
    line<float> input(npoint), output(1), target(npoint);
    SigProc1D sp1;
    // prepare random generetor c++11
    std::random_device rd;
    std::mt19937 e2(rd());
    std::uniform_real_distribution<> dist(-10.f, 10.f);

    for (int i=0 ; i<npoint ; i++)
    {
        input(i) = dist(e2);
    }
    target = input;
    target * value; // scalar * already validated
    sp1.apply_constant_gain(input, output, value);
    EXPECT_EQ(npoint, output.size());
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(target(i), output(i)) << "output[" << i << "] not equal to expected value";
    }

    // internal function
    sp1.edit_raw_data_vector(input);
    sp1.apply_constant_gain(value);
    line<float> internal_vector = sp1.get_filtered_data();
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ(internal_vector(i), output(i)) << "error internal function at loop " << i;
    }

    // external pointers
    float *input_p=nullptr, *output_p=nullptr;
    create_vector(input_p, npoint);
    create_vector(output_p, npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        input_p[i] = input(i);
    }
    sp1.apply_constant_gain(input_p, output_p, npoint, value);
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( output_p[i], output(i) ) << "error external pointers at loop " << i;
    }
    delete_vector(input_p);
    delete_vector(output_p);
}

TEST( Sigproc1D_basic_function, target_gain_interpolation )
{
    int npoint = 3;
    line<float> line1(npoint), target_gain(2), actual_gain(npoint);

    // input gain with size < spi Nel
    target_gain(0) = 0.9f, target_gain(1) = 2.1f;
    SigProc1D sp1;
    sp1.edit_raw_data_vector(line1);
    sp1.edit_internal_gain(target_gain);
    actual_gain = sp1.get_internal_gain();
    EXPECT_FLOAT_EQ( target_gain(0), actual_gain(0) );
    EXPECT_FLOAT_EQ( 1.5f, actual_gain(1) );
    EXPECT_FLOAT_EQ( 2.1f, actual_gain(2) );
    line1 = actual_gain;

    // input gain with the same size than sp1 Nel
    npoint = 666;
    line1.resize(npoint);
    target_gain.resize(npoint);
    sp1.edit_raw_data_vector(line1);
    for (int i=0 ; i<npoint ; i++)
    {
        target_gain(i) = float(i)/0.33f;
    }
    sp1.edit_internal_gain(target_gain);
    actual_gain = sp1.get_internal_gain();
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( target_gain(i), actual_gain(i) ) << i;
    }

    //input gain with size > sp1 Nel
    npoint = 33;
    line1.resize(npoint);
    target_gain.resize(2*npoint-1);
    actual_gain.resize(npoint);
    sp1.edit_raw_data_vector(line1);
    for (int i=0 ; i<2*npoint-1 ; i++)
    {
        target_gain(i) = float(i)*0.1247f;
    }
    sp1.edit_internal_gain(target_gain);
    actual_gain = sp1.get_internal_gain();
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( target_gain(2*i), actual_gain(i) ) << "target failure for i = " << i;
    }

    // If we take a triangle function with N segemnts -> N+1 maxima (positive and negative)
    // If we put P+1 points regulary space on each segment (included the maxima) then we have:
    // - P*N+1 total points
    // - if N is even, sum of y(i) = N/2*mean(segment) + y(last point), for each P>=1
    // test with mean = 0, maxima = +/1
    int N = 10;
    int P = 0;
    int target_point;
    for (P=500 ; P<10001 ; P+=500)
    {
        npoint = N+1;
        target_point = N*P+1;
        line1.resize(target_point);
        target_gain.resize(target_point);
        actual_gain.resize(npoint);
        for (int i=0 ; i<npoint ; i++)
        {
            actual_gain(i) = 4.f*pow(-1.f, i);
        }
        sp1.edit_raw_data_vector(line1);
        sp1.edit_internal_gain(actual_gain);
        target_gain = sp1.get_internal_gain();
        //not too greedy on relative error due to number of rounded value
        EXPECT_LE( relative_error(actual_gain(actual_gain.size()-1), target_gain.sum()), 1e-2 ) << "P = " << P;
    }

    // oversampling then undersampling
    line<float> initial_gain;
    N = 4;
    npoint = N+1;
    P = 2;
    target_point = N*P+1;
    line1.resize(target_point);
    actual_gain.resize(npoint);
    for (int i=0 ; i<npoint ; i++)
    {
        actual_gain(i) = 2.589f*pow(-1.f, i);
    }
        // first sampling
    sp1.edit_raw_data_vector(line1);
    sp1.edit_internal_gain(actual_gain);
    initial_gain = sp1.get_internal_gain();
        // oversampling
    P = 20;
    target_point = N*P+1;
    line1.resize(target_point);
    sp1.edit_raw_data_vector(line1);
    sp1.edit_internal_gain(initial_gain);
    actual_gain = sp1.get_internal_gain();
        // undersampling
    P = 2;
    target_point = N*P+1;
    line1.resize(target_point);
    sp1.edit_raw_data_vector(line1);
    sp1.edit_internal_gain(actual_gain);
    target_gain = sp1.get_internal_gain();
    for (int i=0 ; i<initial_gain.size() ; i++)
    {
        EXPECT_LE( relative_error(initial_gain(i),target_gain(i)), 1e-4) << "target problem: i = " << i;;
        // once again not to greedy on relative error due to interpolation
    }

    // external pointers interpolation will be tested via apply_external_gain
}

TEST( Sigproc1D_basic_function, apply_gain )
{
    // interpolation of the gain vector has been validated, vectorial multiplication also
    // so we make just a simple test with a linear function
    int npoint = 149;
    float gradient = 14.6f, shift = -23.23f;
    line<float> target(npoint), ext_gain(2);
    for (int i=0 ; i<npoint ; i++)
    {
        target(i) = gradient * float(i) + shift;
    }
    ext_gain(0) = target(0);
    ext_gain(1) = target(npoint-1);
    SigProc1D sp1;
    sp1.edit_raw_data_vector(target);
    sp1.edit_internal_gain(ext_gain);
    sp1.apply_gain();
    line<float> output = target*target;
    line<float> filt_data = sp1.get_filtered_data();
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( output(i), filt_data(i) ) << "compare apply gain loop " << i;
    }
}

TEST( Sigproc1D_basic_function, external_gain )
{
    int npoint = 471;
    int npoint_gain = 15;
    line<float> input_line(npoint), output_line1, output_line2, gain_line(npoint_gain);
    SigProc1D sp1;

    // RNG
    std::random_device rd;
    std::mt19937 e2(rd());
    std::uniform_real_distribution<> dist(-10.f, 10.f);
 
    for (int i=0 ; i<npoint_gain ; i++)
    {
        gain_line(i) = dist(e2);
    }
    for (int i=0 ; i<npoint ; i++)
    {
        input_line(i) = std::sin(float(i)/10.f);
    }

    sp1.edit_raw_data_vector(input_line);
    sp1.edit_internal_gain(gain_line);
    sp1.apply_gain();
    output_line1 = sp1.get_filtered_data();
    sp1.reset_filtered_data_vector();

    sp1.apply_external_gain(gain_line);
    output_line2 = sp1.get_filtered_data();
    sp1.reset_filtered_data_vector();

    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( output_line2(i), output_line1(i) ) << "internal line faillure at loop " << i;
    }

    // line function
    sp1.apply_external_gain( input_line, output_line2, gain_line );
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( output_line2(i), output_line1(i) ) << "external line base faillure at loop " << i;
    }

    // pointer function
    float *input_p=nullptr, *output_p=nullptr, *gain_p=nullptr;
    create_vector(input_p, npoint);
    create_vector(output_p, npoint);
    create_vector(gain_p, npoint_gain);
    for (int i=0 ; i<npoint_gain ; i++)
    {
        gain_p[i] = gain_line(i);
    }
    for (int i=0 ; i<npoint ; i++)
    {
        input_p[i] = input_line(i);
    }
    sp1.apply_external_gain(input_p, output_p, npoint, gain_p, npoint_gain);
    for (int i=0 ; i<npoint ; i++)
    {
        EXPECT_FLOAT_EQ( output_p[i], output_line1(i) ) << "pointer base faillure at loop " << i;
    }
    delete_vector(gain_p);
    delete_vector(output_p);
    delete_vector(input_p);
}

int main(int argc, char **argv)
{
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
